#
# Author:: Noah Kantrowitz <noah@coderanger.net>
#
# Copyright 2014, Noah Kantrowitz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import gzip
import os
import re

import pytest
import six

from depot.yum import YumRepoMD, YumPrimary, YumFileLists, YumOther


# Convert XML into a format that diffs nicely
def unify_spacing(data):
    if not isinstance(data, six.binary_type):
        data._fileobj.seek(0, 0)
        data = data._fileobj.read()
    return re.sub('>', '>\n', re.sub('\s+', ' ', re.sub('>', '>\n', re.sub('\n', '', data))))

def fixture(cls, name):
    path = os.path.join(os.path.dirname(__file__), 'data', name)
    open_ = gzip.open if path.endswith('.gz') else open
    fileobj = open_(path, 'rb')
    obj =  cls.from_file(path, fileobj=fileobj)
    obj._fileobj = fileobj
    return obj


class TestYumRepoMD(object):
    @pytest.fixture
    def epel(self):
        return fixture(YumRepoMD, 'epel_repomd.xml')

    @pytest.fixture
    def pgdg(self):
        return fixture(YumRepoMD, 'pgdg_repomd.xml')

    def test_from_file(self, epel):
        assert epel.revision == 1389466441
        assert epel.tags == ['binary-i386']
        assert epel['filelists']['checksum'] == '8892c3e34eef269cea677558ee5a40057052ecff'
        assert epel['filelists']['open-checksum'] == 'b6081d9789bf5cef2ffac8c03fa67224b8c22f53'
        assert epel['group']['location'] == 'repodata/fdddf90da3a700ad6da5ff78e13c17258655bbe3-comps-el5.xml'

    def test_str_epel(self, epel):
        assert unify_spacing(epel.encode()) == unify_spacing(epel)

    def test_str_pgdg(self, pgdg):
        assert unify_spacing(pgdg.encode()) == unify_spacing(pgdg)


class TestYumPrimary(object):
    @pytest.fixture
    def epel(self):
        return fixture(YumPrimary, 'epel_primary.xml.gz')

    @pytest.fixture
    def pgdg(self):
        return fixture(YumPrimary, 'pgdg_primary.xml')

    @pytest.fixture
    def pgdgmini(self):
        return fixture(YumPrimary, 'pgdgmini_primary.xml')

    def test_from_file_pgdgmini(self, pgdgmini):
        assert pgdgmini[('ip4r93', 'x86_64', '0:1.05-3.rhel6')]['summary'] == 'IPv4 and IPv4 range index types for PostgreSQL'
        assert pgdgmini[('ip4r93', 'x86_64', '0:1.05-3.rhel6')]['url'] == 'http://pgfoundry.org/projects/ip4r'
        assert pgdgmini[('postgresql93-debuginfo', 'x86_64', '0:9.3.1-1PGDG.rhel6')]['checksum'] == '048be8c5573c0d92bf64e476beb739c31b2d0f91'
        assert pgdgmini[('postgresql93-debuginfo', 'x86_64', '0:9.3.1-1PGDG.rhel6')]['url'] == ''

    def test_str_epel(self, epel):
        assert unify_spacing(epel.encode()) == unify_spacing(epel)

    def test_str_pgdg(self, pgdg):
        assert unify_spacing(pgdg.encode()) == unify_spacing(pgdg)

    def test_str_pgdgmini(self, pgdgmini):
        assert unify_spacing(pgdgmini.encode()) == unify_spacing(pgdgmini)


class TestYumFileLists(object):
    @pytest.fixture
    def epel(self):
        return fixture(YumFileLists, 'epel_filelists.xml.gz')

    @pytest.fixture
    def pgdg(self):
        return fixture(YumFileLists, 'pgdg_filelists.xml')

    def test_str_epel(self, epel):
        assert unify_spacing(epel.encode()) == unify_spacing(epel)

    def test_str_pgdg(self, pgdg):
        assert unify_spacing(pgdg.encode()) == unify_spacing(pgdg)


class TestYumOther(object):
    @pytest.fixture
    def epel(self):
        return fixture(YumOther, 'epel_other.xml.gz')

    @pytest.fixture
    def pgdg(self):
        return fixture(YumOther, 'pgdg_other.xml')

    def test_str_epel(self, epel):
        assert unify_spacing(epel.encode()) == unify_spacing(epel)

    def test_str_pgdg(self, pgdg):
        assert unify_spacing(pgdg.encode()) == unify_spacing(pgdg)



#
# Author:: Noah Kantrowitz <noah@coderanger.net>
#
# Copyright 2014, Noah Kantrowitz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import collections

from defusedxml import lxml
from lxml.builder import ElementMaker
from lxml.etree import QName
import six


class YumMeta(collections.OrderedDict):
    nsmap = {}

    def __init__(self, *args, **kwargs):
        self.filename = kwargs.pop('filename') if 'filename' in kwargs else None
        super(YumMeta, self).__init__(*args, **kwargs)

    @classmethod
    def from_file(cls, filename=None, fileobj=None, *args, **kwargs):
        fileobj = fileobj or open(filename, 'rb')
        kwargs['filename'] = filename
        kwargs['root'] = lxml.parse(fileobj)
        return cls.from_element(*args, **kwargs)

    @classmethod
    def from_element(cls, root, *args, **kwargs):
        raise NotImplementedError

    def to_element(self, E):
        raise NotImplementedError

    def encode(self):
        return lxml.tostring(
            self.to_element(ElementMaker(nsmap=self.nsmap)),
            xml_declaration=True,
            encoding='UTF-8',
        )


class YumData(collections.OrderedDict):
    NS = None
    # Sentinel object
    NoInsert = object()

    @classmethod
    def from_element(cls, root):
        self = cls(**cls.root_from_element(root))
        for elm in root.findall('*'):
            key = QName(elm.tag).localname
            fn = getattr(self, '{0}_from_element'.format(key.replace('-', '_')), None)
            val = fn(key, elm) if fn else elm.text
            if val is not self.NoInsert:
                self[key] = val
        return self

    @classmethod
    def root_from_element(cls, root):
        return root.attrib

    def to_element(self, E):
        sub = []
        for key, value in six.iteritems(self):
            ns_key = '{{{0}}}{1}'.format(self.NS, key) if self.NS else key
            fn = getattr(self, '{0}_to_element'.format(key.replace('-', '_')), None)
            if fn:
                elm = fn(E, ns_key, value)
            elif value:
                elm = E(ns_key, value)
            else:
                elm = E(ns_key)
            sub.append(elm)
        return self.root_to_element(E, sub)

    def root_to_element(self, E, sub):
        return E.data(*sub)



#
# Author:: Noah Kantrowitz <noah@coderanger.net>
#
# Copyright 2014, Noah Kantrowitz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import six

from .base import YumMeta, YumData


class YumRepoMDData(YumData):
    def __init__(self, type, *args, **kwargs):
        super(YumRepoMDData, self).__init__(*args, **kwargs)
        self.type = type

    def location_from_element(self, key, elm):
        return elm.attrib['href']

    def location_to_element(self, E, key, value):
        return E(key, href=value)

    def checksum_to_element(self, E, key, value):
        return E(key, value, type='sha')
    open_checksum_to_element = checksum_to_element

    def root_to_element(self, E, sub):
        return E.data(*sub, type=self.type)


class YumRepoMD(YumMeta):
    DataClass = YumRepoMDData
    nsmap = {
        None: 'http://linux.duke.edu/metadata/repo',
        'rpm': 'http://linux.duke.edu/metadata/rpm',
    }

    def __init__(self, revision=0, tags=None, *args, **kwargs):
        super(YumRepoMD, self).__init__(*args, **kwargs)
        self.revision = revision
        self.tags = tags or []

    @classmethod
    def from_element(cls, root, *args, **kwargs):
        kwargs['revision'] = int(root.find('{*}revision').text)
        kwargs['tags'] = [elm.text for elm in root.findall('{*}tags/{*}content')]
        self = cls(*args, **kwargs)
        for elm in root.findall('{*}data'):
            data = self.DataClass.from_element(elm)
            self[data.type] = data
        return self

    def to_element(self, E):
        sub = [E.revision(str(self.revision))]
        if self.tags:
            sub.append(E.tags(*[E.content(tag) for tag in self.tags]))
        for data in six.itervalues(self):
            sub.append(data.to_element(E))
        return E.repomd(*sub)



#
# Author:: Noah Kantrowitz <noah@coderanger.net>
#
# Copyright 2014, Noah Kantrowitz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

from depot.storage import StorageWrapper

class TestStorage(object):
    def test_file_abspath(self):
        fileobj = StorageWrapper.file(__file__)
        assert isinstance(fileobj, file)

    def test_file_relpath(self):
        cwd = os.getcwd()
        try:
            os.chdir(os.path.dirname(__file__))
            fileobj = StorageWrapper.file('test_storage.py')
            assert isinstance(fileobj, file)
        finally:
            os.chdir(cwd)

    def test_file_uri(self):
        # Should really just poke the dummy driver here instead
        class StorageWrapperTest(StorageWrapper):
            def download_iter(self, path, *args, **kwargs):
                return (path,)
        fileobj = StorageWrapperTest.file('dummy://bucket/path/to/txt')
        assert isinstance(fileobj, file)
        assert fileobj.read() == 'path/to/txt'



import hashlib
import os
import tempfile

import six
import libcloud.security
from libcloud.storage.providers import get_driver
from libcloud.storage.types import ContainerDoesNotExistError, ObjectDoesNotExistError
from six.moves.urllib.parse import urlparse

# Include the current cURL CA bundle as a fallback
_base_path = os.path.abspath(os.path.dirname(__file__))
libcloud.security.CA_CERTS_PATH.append(os.path.join(_base_path, 'cacert.pem'))


class Sizer(object):
    def __init__(self):
        self.size = 0

    def update(self, data):
        self.size += len(data)


class StorageWrapper(object):
    def __init__(self, uri, no_public=False):
        self.uri = urlparse(uri)
        self.no_public = no_public
        self.storage = self._get_storage(self.uri)
        self._hashes = {}

    def upload(self, path, data):
        if isinstance(data, six.binary_type):
            # Simple case of in-memory content
            return self.upload(path, (data,))
        elif hasattr(data, 'read'):
            # File-like object
            def it(size=4096):
                buf = data.read(size)
                while buf:
                    yield buf
                    buf = data.read(size)
            return self.upload(path, it())
        else:
            # Just try iterating
            self._update_hashes(path, '')
            path_root, path_ext = os.path.splitext(path)
            extra = {}
            if path_ext == '.gz':
                extra['content_type'] = 'application/x-gzip'
            else:
                extra['content_type'] = 'text/plain'
            if not self.no_public:
                if self.uri.scheme.startswith('s3'):
                    extra['acl'] = 'public-read'
                # Other clouds here
            return self.storage.upload_object_via_stream((self._update_hashes(path, buf, False) for buf in data), path, extra=extra)

    def download(self, path, skip_hash=False):
        # Assumption, this isn't a big file
        it = self.download_iter(path, skip_hash)
        if it:
            return ''.join(it)

    def download_iter(self, path, skip_hash=False):
        try:
            obj = self.storage.get_object(path)
        except ObjectDoesNotExistError:
            return None
        if not skip_hash:
            self._update_hashes(path, '')
        return (
            buf if skip_hash else self._update_hashes(path, buf, False)
            for buf in self.storage.download_object_as_stream(obj)
        )

    def __contains__(self, path):
        try:
            self.storage.get_object(path)
            return True
        except ObjectDoesNotExistError:
            return False

    def _update_hashes(self, path, data, reset=True):
        if reset or path not in self._hashes:
            self._hashes[path] = {
                'md5': hashlib.md5(),
                'sha1': hashlib.sha1(),
                'sha256': hashlib.sha256(),
                'size': Sizer(),
            }
        if data:
            for hasher in six.itervalues(self._hashes[path]):
                hasher.update(data)
        return data

    def hashes(self, path):
        if path not in self._hashes:
            self._update_hashes(path, '')
        return self._hashes[path]

    @classmethod
    def file(cls, uri_or_path):
        """
        Given either a URI like s3://bucket/path.txt or a path like /path.txt,
        return a file object for it.
        """
        uri = urlparse(uri_or_path)
        if not uri.scheme:
            # Just a normal path
            return open(uri_or_path, 'rb')
        else:
            it = cls(uri_or_path).download_iter(uri.path.lstrip('/'), skip_hash=True)
            if not it:
                raise ValueError('{0} not found'.format(uri_or_path))
            tmp = tempfile.TemporaryFile()
            for chunk in it:
                tmp.write(chunk)
            tmp.seek(0, 0)
            return tmp

    @classmethod
    def _get_storage(cls, uri):
        """
        Given a URI like local:///srv/repo or s3://key:secret@apt.example.com,
        return a libcloud storage or container object.
        """
        driver = cls._get_driver(uri.scheme)
        key = uri.username
        secret = uri.password
        container = uri.netloc
        driver_kwargs = {}
        if uri.scheme.startswith('s3'):
            if not key:
                key = os.environ.get('AWS_ACCESS_KEY_ID')
            if not secret:
                secret = os.environ.get('AWS_SECRET_ACCESS_KEY')
            if not (key and secret and container):
                raise ValueError('For S3 you must provide an access key ID, secret access key, and bucket name')
            # No way to store this in the URI, what about a CLI option too?
            if 'AWS_TOKEN' in os.environ:
                driver_kwargs['token'] = os.environ['AWS_TOKEN']
        elif uri.scheme == 'local':
            parts = []
            if uri.netloc:
                parts.append(uri.netloc)
            if uri.path:
                parts.append(uri.path)
            if not parts:
                parts.append('.')
            base_path = os.path.abspath(''.join(parts))
            key = os.path.dirname(base_path)
            container = os.path.basename(base_path)
        storage = driver(key, secret, **driver_kwargs)
        try:
            return storage.get_container(container)
        except ContainerDoesNotExistError:
            return storage.create_container(container)

    @classmethod
    def _get_driver(cls, name):
        """Wrapper for libcloud's get_driver for testing."""
        return get_driver(name)



import getpass
import os

import gnupg


class GPG(object):
    def __init__(self, keyid):
        self.gpg = gnupg.GPG(use_agent=False)
        self.keyid = keyid
        if not self.keyid:
            # Compat with how Freight does it.
            self.keyid = os.environ.get('GPG')
        self.passphrase = None
        self._verify()

    def _verify(self):
        """Some sanity checks on GPG."""
        if not self.keyid:
            raise ValueError('No GPG key specified for signing, did you mean to use --no-sign?')
        sign = self.gpg.sign('', keyid=self.keyid)
        if 'secret key not available' in sign.stderr:
            raise ValueError('Key not found')
        elif 'NEED_PASSPHRASE' in sign.stderr:
            self.passphrase = getpass.getpass('Passphrase for GPG key: ')

    def sign(self, data, detach=False):
        sign = self.gpg.sign(data, keyid=self.keyid, passphrase=self.passphrase, detach=detach)
        if not sign:
            raise ValueError(sign.stderr)
        return str(sign)

    def public_key(self):
        return self.gpg.export_keys(self.keyid)



import os
import hashlib

import pytest
from pretend import call_recorder, call, stub

from depot.apt import AptPackages, AptRelease, AptRepository

def fixture_path(*path):
    return os.path.join(os.path.dirname(__file__), 'data', *path)

@pytest.fixture
def storage():
    return stub(
        hashes=call_recorder(lambda path: {'size': stub(size=1), 'sha1': hashlib.sha1(), 'sha256': hashlib.sha256(), 'md5': hashlib.md5()}),
        __contains__=lambda key: False,
        upload=lambda path, fileobj: None,
    )

class TestAptPackages(object):
    @pytest.fixture
    def pgdg(self, storage):
        data = open(os.path.join(os.path.dirname(__file__), 'data', 'pgdg_Packages'), 'rb').read()
        return AptPackages(storage, data)

    def test_reading(self, pgdg):
        # Check that the keys match
        packages = sorted(['libecpg-compat2', 'libecpg-dev', 'libecpg5', 'libpgtypes2', 'libpq-dev', 'libpq5'])
        assert sorted(pgdg.packages.keys()) == [(p, '8.2.23-1.pgdg12.4+1') for p in packages]
        # Just testing a few random keys since an exhaustive test would be too verbose
        assert pgdg.packages[('libecpg-compat2', '8.2.23-1.pgdg12.4+1')]['Filename'] == 'pool/8.2/p/postgresql-8.2/libecpg-compat2_8.2.23-1.pgdg12.4+1_amd64.deb'
        assert pgdg.packages[('libpgtypes2', '8.2.23-1.pgdg12.4+1')]['SHA256'] == 'bc711ca44adfc4425e69a1777c4284ac942b3f352d8296a3047895e97d82489b'
        assert pgdg.packages[('libpq-dev', '8.2.23-1.pgdg12.4+1')]['Description'] == '''header files for libpq5 (PostgreSQL library)
 Header files and static library for compiling C programs to link
 with the libpq library in order to communicate with a PostgreSQL
 database backend.
 .
 PostgreSQL is an object-relational SQL database management system.'''

    def test_adding(self, pgdg):
        package_data = {'Package': 'test', 'Version': '1'}
        package = stub(pool_path='pool/main/t/test/test.deb', __getitem__=package_data.__getitem__, __setitem__=package_data.__setitem__)
        pgdg.add(package)
        assert pgdg.packages[('test', '1')]['Filename'] == 'pool/main/t/test/test.deb'
        assert pgdg.packages[('test', '1')]['Size'] == '1'
        assert pgdg.packages[('test', '1')]['SHA1'] == 'da39a3ee5e6b4b0d3255bfef95601890afd80709'

    def test_serializing(self, pgdg):
        assert str(pgdg).strip() == pgdg._data.strip()


class TestAptRelease(object):
    @pytest.fixture
    def pgdg(self, storage):
        data = open(os.path.join(os.path.dirname(__file__), 'data', 'pgdg_Release'), 'rb').read()
        return AptRelease(storage, 'precise-pgdg', data)

    def test_reading(self, pgdg):
        assert pgdg['Architectures'] == 'amd64 i386'
        assert pgdg['Date'] == 'Mon, 30 Dec 2013 17:34:01 UTC'

    def test_reading_hashes(self, pgdg):
        assert pgdg.hashes['md5']['8.3/source/Release'] == ('0ca3534d616a6a902c6ed3877d30e740', '132')

    def test_serializing(self, pgdg):
        assert str(pgdg).strip() == pgdg._data.strip()

    def test_add_metadata(self, pgdg):
        assert pgdg['Components'] == 'main 8.2 8.3 8.4 9.0 9.1 9.2 9.3'
        assert pgdg['Architectures'] == 'amd64 i386'
        pgdg.add_metadata('test', 'amd64')
        assert pgdg['Components'] == '8.2 8.3 8.4 9.0 9.1 9.2 9.3 main test'
        assert pgdg['Architectures'] == 'amd64 i386'
        pgdg.add_metadata('test', 'test')
        assert pgdg['Components'] == '8.2 8.3 8.4 9.0 9.1 9.2 9.3 main test'
        assert pgdg['Architectures'] == 'amd64 i386 test'

    def test_update_hash(self, pgdg):
        pgdg.update_hash('main/binary-amd64/Packages')
        assert pgdg.storage.hashes.calls == [call('dists/precise-pgdg/main/binary-amd64/Packages')]
        assert pgdg.hashes['md5']['main/binary-amd64/Packages'] == ('d41d8cd98f00b204e9800998ecf8427e', '1')
        assert pgdg.hashes['sha1']['main/binary-amd64/Packages'] == ('da39a3ee5e6b4b0d3255bfef95601890afd80709', '1')
        assert pgdg.hashes['sha256']['main/binary-amd64/Packages'] == ('e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855', '1')

class TestAptRepository(object):
    @pytest.fixture
    def package_path(self):
        return fixture_path('depot_1.2.3-1_amd64.deb')

    def test_duplicate_upload(self, storage, package_path):
        storage.__contains__ = lambda key: True
        repo = AptRepository(storage, None, None)
        assert not repo.add_package(package_path)

    def test_duplicate_upload_force(self, storage, package_path):
        storage.__contains__ = lambda key: True
        repo = AptRepository(storage, None, None)
        assert repo.add_package(package_path, force=True)

    def test_nonduplicate_upload(self, storage, package_path):
        repo = AptRepository(storage, None, None)
        assert repo.add_package(package_path, force=True)



#
# Author:: Noah Kantrowitz <noah@coderanger.net>
#
# Copyright 2014, Noah Kantrowitz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import six

from .base import YumMeta, YumData


class YumFileListsPackage(YumData):
    def __init__(self, pkgid, name, arch, *args, **kwargs):
        super(YumFileListsPackage, self).__init__(*args, **kwargs)
        self.pkgid = pkgid
        self.name = name
        self.arch = arch
        self.files = []

    def version_from_element(self, key, elm):
        return elm.attrib

    def file_from_element(self, key, elm):
        self.files.append((elm.attrib.get('type'), elm.text))
        return self.NoInsert

    def version_to_element(self, E, key, value):
        elm = E(key)
        for a_key, a_value in six.iteritems(value):
            elm.attrib[a_key] = a_value
        return elm

    def root_to_element(self, E, sub):
        for file_type, file_path in self.files:
            elm = E.file(file_path)
            if file_type:
                elm.attrib['type'] = file_type
            sub.append(elm)
        root = E.package(*sub)
        root.attrib['pkgid'] = self.pkgid
        root.attrib['name'] = self.name
        root.attrib['arch'] = self.arch
        return root


class YumFileLists(YumMeta):
    PackageClass = YumFileListsPackage
    nsmap = {None: 'http://linux.duke.edu/metadata/filelists'}

    @classmethod
    def from_element(cls, root, *args, **kwargs):
        self = cls(*args, **kwargs)
        for elm in root.findall('{*}package'):
            pkg = self.PackageClass.from_element(elm)
            self[pkg.pkgid] = pkg  # Should this be a (name, arch, ver) tuple like YumPrimary?
        return self

    def to_element(self, E):
        return E.filelists(*[pkg.to_element(E) for pkg in six.itervalues(self)], packages=str(len(self)))



"""Usage: depot [options] <package> [<package> ...]

-h --help                    show this help message and exit
--version                    show program's version number and exit
-s URI --storage=URI         URI for storage provider, checks $DEPOT_STORAGE or local://
-c NAME --codename=NAME      Debian distribution codename [default: lucid]
--component=NAME             Debian component name [default: main]
-a ARCH --architecture=ARCH  package architecture if not specified in package
-k KEYID --gpg-key=KEYID     GPG key ID to use for signing
--no-sign                    do not sign this upload
--no-public                  do not make cloud files public-readable
--force                      force upload, even if overwriting

Example:
depot -s s3://apt.example.com -c precise -k 6791B14F mypackage.deb

"""

import os

import docopt

from .apt import AptRepository
from .gpg import GPG
from .storage import StorageWrapper
from .version import __version_info__, __version__  # noqa


def main():
    args = docopt.docopt(__doc__, version='depot '+__version__)
    if not args['--storage']:
        args['--storage'] = os.environ.get('DEPOT_STORAGE', 'local://')
    if args['--no-sign']:
        gpg = None
    else:
        gpg = GPG(args['--gpg-key'])
    storage = StorageWrapper(args['--storage'], args['--no-public'])
    repo = AptRepository(storage, gpg, args['--codename'], args['--component'], args['--architecture'])
    for pkg_path in args['<package>']:
        if '@' in pkg_path:
            print('Copying package {0}'.format(pkg_path))
            repo.copy_package(pkg_path)
        else:
            print('Uploading package {0}'.format(pkg_path))
            fileobj = StorageWrapper.file(pkg_path)
            if not repo.add_package(pkg_path, fileobj, args['--force']):
                print('{0} already uploaded, skipping (use --force to override)'.format(pkg_path))
            fileobj.close()
    print('Uploading metadata')
    repo.commit_metadata()



# coding=utf8
import bz2
import collections
import os
import re
import tarfile
import time

try:
    import lzma
except ImportError:
    lzma = None

import arpy
import six

from .utils import gzip_compress
from .version import __version__


class AptMeta(collections.OrderedDict):
    def __init__(self, data):
        super(AptMeta, self).__init__()
        self._data = data  # For testing/debugging
        last_key = None
        for line in data.splitlines() if hasattr(data, 'splitlines') else data.readlines():
            line = line.rstrip('\n')
            if line[0].isspace():
                if last_key:
                    self[last_key] += '\n' + line
                else:
                    raise ValueError('Can not parse line: {0}'.format(repr(line)))
            else:
                last_key, value = line.split(':', 1)
                value = value.lstrip(' ')
                self[last_key] = value

    def __str__(self):
        return '\n'.join('{0}:{1}{2}'.format(key, '' if value[0] == '\n' else ' ', value) for key, value in six.iteritems(self) if value)


class AptPackage(AptMeta):
    def __init__(self, filename, fileobj=None, data=None):
        self.name = filename
        if not data:
            if fileobj:
                self.ar = arpy.Archive(filename or getattr(fileobj, 'name', None), fileobj)
            else:
                self.ar = arpy.Archive(filename)
            self.ar.read_all_headers()
            self.control_tar = tarfile.open('control.tar.gz', 'r:gz', fileobj=self.ar.archived_files['control.tar.gz'])
            data = self.control_tar.extractfile('./control')
        super(AptPackage, self).__init__(data)

    @property
    def pool_path(self):
        filename = self.name or self['Filename']
        return 'pool/{0}'.format(os.path.basename(filename))


class AptPackages(object):
    def __init__(self, storage, data):
        self.storage = storage
        self.packages = {}
        self._data = data  # For testing/debugging
        for buf in data.split('\n\n'):
            if not buf.strip():
                continue
            pkg = AptPackage(None, data=buf)
            self.packages[(pkg['Package'], pkg['Version'])] = pkg

    def add(self, pkg):
        hashes = self.storage.hashes(pkg.pool_path)
        pkg['Filename'] = pkg.pool_path
        pkg['Size'] = str(hashes['size'].size)
        pkg['MD5sum'] = hashes['md5'].hexdigest()
        pkg['SHA1'] = hashes['sha1'].hexdigest()
        pkg['SHA256'] = hashes['sha256'].hexdigest()
        self.packages[(pkg['Package'], pkg['Version'])] = pkg

    def __str__(self, extra_fn=None):
        return '\n\n'.join(str(pkg) for key, pkg in sorted(six.iteritems(self.packages), key=lambda k: k[0]))


class AptRelease(AptMeta):
    def __init__(self, storage, codename, *args, **kwargs):
        self.storage = storage
        self.codename = codename
        super(AptRelease, self).__init__(*args, **kwargs)
        if 'Components' not in self:
            # Need to setup some defaults
            self['Origin'] = 'Depot {0}'.format(__version__)
            self['Date'] = ''  # Will be regenerated, but lock the order
            self['Codename'] = self.codename
            # These are filled in using add_metadata()
            self['Architectures'] = ''
            self['Components'] = ''

        self.hashes = {
            'md5': self._parse_hashes('MD5Sum'),
            'sha1': self._parse_hashes('SHA1'),
            'sha256': self._parse_hashes('SHA256'),
        }

    def _parse_hashes(self, key):
        hashes = collections.OrderedDict()
        if key in self:
            for line in self[key].splitlines():
                line = line.strip()
                if not line:
                    continue
                hash, size, path = line.split(' ', 2)
                hashes[path] = (hash, size)
        return hashes

    def _compile_hashes(self, key):
        return '\n' + '\n'.join(' {0} {1} {2}'.format(hash, size, path) for path, (hash, size) in six.iteritems(self.hashes[key]))

    def update_hash(self, path):
        hashes = self.storage.hashes('dists/{0}/{1}'.format(self.codename, path))
        for hash_type in list(six.iterkeys(self.hashes)):
            self.hashes[hash_type][path] = (hashes[hash_type].hexdigest(), str(hashes['size'].size))

    def add_metadata(self, component, architecture):
        components = set(s for s in self['Components'].split() if s)
        components.add(component)
        self['Components'] = ' '.join(sorted(components))
        architectures = set(s for s in self['Architectures'].split() if s)
        architectures.add(architecture)
        self['Architectures'] = ' '.join(sorted(architectures))

    def __str__(self):
        self['MD5Sum'] = self._compile_hashes('md5')
        self['SHA1'] = self._compile_hashes('sha1')
        self['SHA256'] = self._compile_hashes('sha256')
        if not self.get('Date'):
            now = time.gmtime()
            # The debian standard (Policy 4.4) really does specify the English labels
            day_of_week = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun'][now.tm_wday]
            month = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'][now.tm_mon - 1]
            self['Date'] = time.strftime('{0}, %d {1} %Y %H:%M:%S UTC'.format(day_of_week, month), now)
        return super(AptRelease, self).__str__()


class AptRepository(object):
    COPY_SPEC_RE = re.compile(r'^([\w_-]+)@(.+?)(?:,(.+?))$')

    def __init__(self, storage, gpg, codename, component='main', architecture=None):
        self.storage = storage
        self.gpg = gpg
        self.codename = codename
        self.component = component
        self.architecture = architecture
        self.dirty_packages = {}  # arch: [pkg,+]
        self.dirty_sources = False

    def add_package(self, path, fileobj=None, force=False):
        fileobj = fileobj or open(path, 'rb')
        path = os.path.basename(path)
        pkg = AptPackage(path, fileobj)
        # Check that we have an arch if needed
        arch = pkg['Architecture']
        if pkg['Architecture'] == 'any':
            arch = self.architecture
            if not arch:
                raise ValueError('Architechture required when adding packages for "any"')
        self.dirty_packages.setdefault(arch, []).append(pkg)

        # Check that the package doesn't already exist
        if not force and pkg.pool_path in self.storage:
            return False

        # Stream up the actual package file
        fileobj.seek(0, 0)
        self.storage.upload(pkg.pool_path, fileobj)
        return True

    def copy_package(self, package):
        md = self.COPY_SPEC_RE.match(package)
        if not md:
            raise ValueError('Unable to parse {0)'.format(repr(package)))
        #package_name = md.group(1)
        #package_version = md.group(2)
        #package_component = md.group(3)
        raise NotImplementedError('TODO finish this')

    def commit_package_metadata(self, arch, pkgs):
        # Update the Packages file
        packages_path = 'dists/{0}/{1}/binary-{2}/Packages'.format(self.codename, self.component, arch)
        packages = AptPackages(self.storage, self.storage.download(packages_path, skip_hash=True) or '')
        for pkg in pkgs:
            packages.add(pkg)
        packages_raw = str(packages)
        self.storage.upload(packages_path, packages_raw)
        self.storage.upload(packages_path+'.gz', gzip_compress(packages_raw))
        self.storage.upload(packages_path+'.bz2', bz2.compress(packages_raw))
        if lzma:
            self.storage.upload(packages_path+'.lzma', lzma.compress(packages_raw))

    def commit_sources_metadata(self):
        # Update the Sources file
        sources_path = 'dists/{0}/{1}/source/Sources'.format(self.codename, self.component)
        if sources_path in self.storage:
            return
        sources_content = ''
        self.storage.upload(sources_path, sources_content)
        self.storage.upload(sources_path+'.gz', gzip_compress(sources_content))
        self.storage.upload(sources_path+'.bz2', bz2.compress(sources_content))
        if lzma:
            self.storage.upload(sources_path+'.lzma', lzma.compress(sources_content))
        self.dirty_sources = True

    def commit_release_metadata(self, archs):
        # Update Release
        release_path = 'dists/{0}/Release'.format(self.codename)
        release = AptRelease(self.storage, self.codename, self.storage.download(release_path, skip_hash=True) or '')
        for arch in archs:
            release.add_metadata(self.component, arch)
            release_packages_path = '{0}/binary-{1}/Packages'.format(self.component, arch)
            release.update_hash(release_packages_path)
            release.update_hash(release_packages_path+'.gz')
            release.update_hash(release_packages_path+'.bz2')
            if lzma:
                release.update_hash(release_packages_path+'.lzma')
        if self.dirty_sources:
            release_sources_path = '{0}/source/Sources'.format(self.component)
            release.update_hash(release_sources_path)
            release.update_hash(release_sources_path+'.gz')
            release.update_hash(release_sources_path+'.bz2')
            if lzma:
                release.update_hash(release_sources_path+'.lzma')
        # Force the date to regenerate
        release['Date'] = None
        release_raw = str(release)
        self.storage.upload(release_path, release_raw)

        # GPG signing
        if self.gpg:
            # Fun fact, even debian's own tools don't seem to support this InRelease file
            in_release_path = 'dists/{0}/InRelease'.format(self.codename)
            self.storage.upload(in_release_path, self.gpg.sign(release_raw))
            self.storage.upload(release_path+'.gpg', self.gpg.sign(release_raw, detach=True))

            # Upload the pubkey to be nice
            self.storage.upload('pubkey.gpg', self.gpg.public_key())

    def commit_metadata(self):
        for arch, packages in six.iteritems(self.dirty_packages):
            self.commit_package_metadata(arch, packages)
        self.commit_sources_metadata()
        self.commit_release_metadata(six.iterkeys(self.dirty_packages))



#
# Author:: Noah Kantrowitz <noah@coderanger.net>
#
# Copyright 2014, Noah Kantrowitz
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import six

from .base import YumMeta, YumData


class YumPrimaryFormat(YumData):
    NS = 'http://linux.duke.edu/metadata/rpm'

    def __init__(self, *args, **kwargs):
        super(YumPrimaryFormat, self).__init__(*args, **kwargs)
        self.files = []

    def header_range_from_element(self, key, elm):
        return elm.attrib

    def provides_from_element(self, key, elm):
        return [entry.attrib for entry in elm.findall('{*}entry')]
    requires_from_element = provides_from_element
    conflicts_from_element = provides_from_element
    obsoletes_from_element = provides_from_element

    def file_from_element(self, key, elm):
        self.files.append((elm.attrib.get('type'), elm.text))
        return self.NoInsert

    def vendor_from_element(self, key, elm):
        return elm.text or ''

    def header_range_to_element(self, E, key, value):
        elm = E(key)
        for a_key, a_value in six.iteritems(value):
            elm.attrib[a_key] = a_value
        return elm

    def provides_to_element(self, E, key, value):
        entries = []
        for entry in value:
            e_elm = E('{{{0}}}entry'.format(self.NS))
            for e_key, e_value in six.iteritems(entry):
                e_elm.attrib[e_key] = e_value
            entries.append(e_elm)
        if not entries:
            entries.append('')
        return E(key, *entries)
    requires_to_element = provides_to_element
    conflicts_to_element = provides_to_element
    obsoletes_to_element = provides_to_element

    def root_to_element(self, E, sub):
        for file_type, file_path in self.files:
            elm = E.file(file_path)
            if file_type:
                elm.attrib['type'] = file_type
            sub.append(elm)
        return E.format(*sub)


class YumPrimaryPackage(YumData):
    FormatClass = YumPrimaryFormat

    def __init__(self, type, *args, **kwargs):
        super(YumPrimaryPackage, self).__init__(*args, **kwargs)
        self.type = type

    def location_from_element(self, key, elm):
        return elm.attrib['href']

    def version_from_element(self, key, elm):
        return elm.attrib
    time_from_element = version_from_element
    size_from_element = version_from_element

    def format_from_element(self, key, elm):
        return self.FormatClass.from_element(elm)

    def packager_from_element(self, key, elm):
        return elm.text or ''
    url_from_element = packager_from_element

    def location_to_element(self, E, key, value):
        return E(key, href=value)

    def version_to_element(self, E, key, value):
        elm = E(key)
        for a_key, a_value in six.iteritems(value):
            elm.attrib[a_key] = a_value
        return elm
    time_to_element = version_to_element
    size_to_element = version_to_element

    def checksum_to_element(self, E, key, value):
        elm = E(key, value)
        elm.attrib['type'] = 'sha'
        elm.attrib['pkgid'] = 'YES'
        return elm

    def format_to_element(self, E, key, value):
        return value.to_element(E)

    def packager_to_element(self, E, key, value):
        return E(key, value or '')
    url_to_element = packager_to_element

    def root_to_element(self, E, sub):
        return E.package(*sub, type=self.type)

    def full_version(self):
        return '{epoch}:{ver}-{rel}'.format(**self['version'])


class YumPrimary(YumMeta):
    PackageClass = YumPrimaryPackage
    nsmap = {
        None: 'http://linux.duke.edu/metadata/common',
        'rpm': 'http://linux.duke.edu/metadata/rpm',
    }

    @classmethod
    def from_element(cls, root, *args, **kwargs):
        self = cls(*args, **kwargs)
        for elm in root.findall('{*}package'):
            pkg = self.PackageClass.from_element(elm)
            self[(pkg['name'], pkg['arch'], pkg.full_version())] = pkg
        return self

    def to_element(self, E):
        return E.metadata(*[pkg.to_element(E) for pkg in six.itervalues(self)], packages=str(len(self)))
